#ifndef RSCOMMON_H
#define RSCOMMON_H

#include <sqlite3.h>

#define DEBUG_FLAG_LIST				\
   DF(sql)					\
   DF(sql2)					\
   DF(value)					\
   DF(search)					\
   DF(filter)					\
   DF(check)					\
   DF(lp)

#define debug stdout
#define DEBUG_DEV "/dev/stdout"

#include "common.h"


#define COUNTER_LIST				\
   CTR(commodities_loaded)			\
   CTR(trades_loaded)				\
   CTR(quantities_loaded)			\
   CTR(routes_considered)			\
   CTR(routes_eliminated)			\
   CTR(routes_valued)				\
   CTR(subroute_tails_valued)			\
   CTR(subroutes_valued)			\
   CTR(subroutes_nonempty)
#define CTR(x) extern int ctr_##x;
  COUNTER_LIST
#undef CTR

#define SQL_MUST( call ) ({						 \
    /* `call' is an expression returning result, using  const char *sqe; \
     * chk1 and chk2 are blocks using sqe and  int sqr; */		 \
    const char *sql_must_call_string= #call;				 \
    int sqr;								 \
    if (DEBUGP(sql2)) fprintf(stderr,"SQL %s", sql_must_call_string);	 \
    sqr= (call);							 \
    if (DEBUGP(sql2)) fprintf(stderr," = %d\n", sqr);			 \
    if (sqr) sql_fatal("(unknown)", sqr, sql_must_call_string);		 \
  })									 \

void sql_fatal(const char *stmt_what, int sqr, const char *act_what) NORET;

#define SQL_STEP(ssh) (sql_step((ssh), #ssh, __FILE__, __LINE__))
int sql_step(sqlite3_stmt *ssh, const char *ssh_string,
	     const char *file, int line);

#define SQL_DISTINCT_DECL(cols, nintcols)	\
  int cols[nintcols];				\
  cols[0]= -1;
#define SQL_DISTINCT_STEP(ssh, cols, nkeycols)				 \
  (sql_step_distinct((ssh), #ssh, __FILE__, __LINE__,			 \
		     (cols), sizeof((cols))/sizeof((cols)[0]), nkeycols))
int sql_step_distinct(sqlite3_stmt *ssh, const char *ssh_string,
		      const char *file, int line,
		      int *cols, int ncols, int nkeycols);
   /* These work if we're making a query whose columns consist of:
    *  - keys: integer column(s) on which the results are sorted by the query
    *  - consequences: zero or more integer cols strictly dependent on the keys
    *  - extra: zero or more further (possibly non-integer) columns
    *
    * Call SQL_DISTINCT_DECL, passing intcols = the total number of keys and
    * consequences; it will declare  int cols[intfields];
    *
    * Then each SQL_DISTINCT_STEP is like SQL_STEP only you have to
    * pass the number of key columns and it only returns rows with
    * distinct keys.  Rows with all-identical keys are asserted to
    * have identical consequences.  After each call to
    * SQL_DISTINCT_STEP the keys and consequences will be stored in
    * cols.
    */

int sql_single_int(const char *stmt);

#define SQL_PREPARE(ss,stmt) ((ss)= sql_prepare((stmt),#ss))
sqlite3_stmt *sql_prepare(const char *stmt, const char *what);

#define SQL_BIND(ss,index,value) (sql_bind((ss),(index),(value),#ss,#value))
void sql_bind(sqlite3_stmt *ss, int index, int value,
	      const char *ss_what, const char *val_what);

#define MAX_ROUTELEN 20

extern sqlite3 *db;

void setup_sql(void);


typedef struct {
  double distance_loss_factor;
  struct TradesBlock *trades;
  double route_tail_value;
} IslandPair;

IslandPair *ipair_get_maybe(int si, int di);

double value_route(int nislands, const int *islands, int exclude_arbitrage);
void setup_value(void);

typedef struct {
  double absolute, perleague;
  int absolute_ports[MAX_ROUTELEN], perleague_ports[MAX_ROUTELEN];
} PotentialResult;

void setup_search(void);
void search(int start_isle, PotentialResult ****strat_base_io
                                /* strat_base[finalarch][midarch]-> */);

extern double max_mass, max_volu, max_capi;
extern double distance_loss_factor_per_league;
extern int max_dist;

#define LOSS_FACTOR_PER_DELAY_SLOT (1-1e-8)

extern int islandtablesz;

extern int narches;
extern char **archnames;
extern int *islandid2arch;


#define ONDEMAND(pointer_lvalue, calloc_size_count)			     \
  ((pointer_lvalue) ? :							     \
   ((pointer_lvalue) = mcalloc(sizeof(*(pointer_lvalue)) * calloc_size_count)))


#endif /*RSCOMMON_H*/
